/*-
 * Copyright (C)2014..2018 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2014..2018 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: recobe.c,v 1.38 2019/03/05 13:48:45 babolo Exp $"

#define BLIN_COMPAT     3
#define MIFE_COMPAT     5
#define Bpars_COMPAT    3
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>
#include "recobe.h"

recobe_conf   recobe_cf;

int  recobe_TestOrRevert __P((const char *p));
int  recobe_PrintAndMake __P((const char *p));
int  recobe_PrintNoRecod __P((const char *p));
int  recobe_WriteAndMake __P((const char *p));
int  recobe_G711DiffConv __P((const char *p));
int  recobe_PrintFilling __P((const char *p));
int  recobe_InterMediaTo __P((const char *p));
int  recobe_EtalonCompar __P((const char *p));
int  recobe_ReCalculator __P((const char *p));
int  recobe_ReCoordinate __P((const char *p));
int  recobe_PrintWideStr __P((const char *p));
void recobe_p256         __P(( const char *template
                             , const char *from
                             , const char *to
                             , u_int16_t domain
                             , const u_char array[256]
                            ));

static void
/*****************************************************************************************************/
usage(int ex) {                                                                                  /****
 *****************************************************************************************************/
    fprintf( stderr
           , "recobe @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "Usage: recobe [-options][ infile_or_string[ oufile]]\n"
             "   -b TEXT         mark before unconverted sym, default \"?\"\n" 
             "   -c XY           translate byte-to-byte infile to oufile, X & Y list at -? c\n"
             "   -e TEXT         mark after unconverted sym, default empty\n"
             "   -E -X XX        use XX as env name for opt's arg\n"
             "   -f              full control log\n"
             "   -H              substXXXXsubst substitute if conversion error XXXX\n"
             "   -k cnv          convert text-to-text infile to oufile, cnv list at -? k\n"
             "   -l              ?\n"
             "   -m TEXT         next conversion parameter (TB - str len)(Q -M)\n"
             "   -M SYM          hex convert start symbol\n"
             "   -o ord          byte orer: B, L\n"
             "   -p              print printables as symbols\n"
             "   -P              fancy print\n"
             "   -q              silent\n"
             "   -s code         control code, list at -? s\n"
             "   -S              strict recoding\n"
             "   -t XY[name]     print table from binary stdin\n"
             "   -T              tracing\n"
             "   -v[v[v...]]     verbose\n"
             "   -z TEXT         parce TEXT as arguments continuation\n"
             "   -Z TEXT         parce TEXT as arguments continuation with env substitution\n"
             "   -= mode         internal diagnostic\n"
             "      mode: 0                 blin_ctl(BLIN_CTL_DUMP)\n"
             "            1                 babolo_optest(OptionsString)\n"
             "            2                 babolo_optext(OptionsDescriptor)\n"
             "            3                 babolo_dumpopts(OptionsDescriptor)\n"
             "            4[D]              inconsistent reverse, D - domain (1 TEXT, 2 G711)\n"
             "            5[D[Y[Z]]]        print table(s); D - domain, Y, Z code names\n"
             "            S[D[Y[Z]]]        write table(s); D - domain, Y, Z code names\n"
             "            6[A[B[name]]]     print norecoding table nameAB[]\n"
             "            7{u|A|Au|uA}      inconsistent G711u, G711A, A=>u, u=>a\n"
             "            8[D]              count, D - domain\n"
             "            9[Y[Z]]           test intermediators; Y, Z code names\n"
             "            E[Y[Z]]           compare with etalons\n"
             "            Rk[w[a[i[m[z]]]]] tests\n"
             "            Ck[w[a[i[m[z]]]]] create full recode tables from ETAT domain\n"
             "            W                 make and print raw recobe_1k[65536]\n"
             "            WC                make and print recobe_ud[256]\n"
             "   -?              this text\n"
             "   -?c             domains and encodings list\n"
             "   -?k             k conversions list\n"
             "   -?s             control sums list\n"
           )
    ;
    exit(ex);
}

static void
/*****************************************************************************************************/
usagec(int ex) {                                                                                  /****
 *****************************************************************************************************/
    int               i, k, m, n, u;

    for (i = 0; !!recobe_collection.domains[i]; i++) {
        m = 0;
        for (k = 0; k < recobe_collection.domains[i]->recountad; ++k) {
            if  (!k) fprintf(stderr, "%s", recobe_collection.domains[i]->domname);
            if  (  !!recobe_collection.domains[i]->encolist[k].encoid
                && !!recobe_collection.domains[i]->encolist[k].enconamess
                ) {
                fprintf( stderr
                       , "\t%c%s"
                       , (!m) ? ':' : ' '
                       , recobe_collection.domains[i]->encolist[k].encoid
                       )
                ;
                ++m;
                u = 0;
                for (n = 0; !!(*recobe_collection.domains[i]->encolist[k].enconamess)[n]; ++n) {
                    fprintf( stderr
                           , "%c %s"
                           , (!u) ? ' ' : ','
                           , (*recobe_collection.domains[i]->encolist[k].enconamess)[n]
                           )
                    ;
                    ++u;
                }
                fprintf(stderr, "\n");
    }   }   }
    exit(ex);
}

static void
/*****************************************************************************************************/
usages(int ex) {                                                                                  /****
 *****************************************************************************************************/
    fprintf( stderr
           , "   1  csum1 begin to end\n"
             "  -1  csum1 end to begin\n"
             "   2  crc24 begin to end\n"
             "  -2  crc24 end to begin\n"
             "   3  crc32 begin to end\n"
             "  -3  crc32 end to begin\n"
           )
    ;
    exit(ex);
}

static void
/*****************************************************************************************************/
usagek(int ex) {                                                                                  /****
 *****************************************************************************************************/
    fprintf( stderr
           , "   u0  utf-8 to 1 byte long int\n"
             "   u1  utf-8 to 2 byte long int\n"
             "   u2  utf-8 to 3 byte long int\n"
             "   u3  utf-8 to 4 byte long int\n"
             "   u4  utf-8 to 5 byte long int\n"
             "   u5  utf-8 to 6 byte long int\n"
             "   u6  utf-8 to 7 byte long int\n"
             "   u7  utf-8 to 8 byte long int\n"
             "   u8  utf-8 to 9 byte long int\n"
             "   u9  utf-8 to 10 byte long int\n"
             "   0k  1 byte long int to koi8\n"
             "   1k  2 byte long int to koi8\n"
             "   2k  3 byte long int to koi8\n"
             "   3k  4 byte long int to koi8\n"
             "   4k  5 byte long int to koi8\n"
             "   5k  6 byte long int to koi8\n"
             "   6k  7 byte long int to koi8\n"
             "   7k  8 byte long int to koi8\n"
             "   8k  9 byte long int to koi8\n"
             "   9k 10 byte long int to koi8\n"
           )
    ;
    exit(ex);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
main(int argc, char **argv) {                                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
    const char       *flar = "c:e:fHk:lm:M:o:pPqs:St:vwW=:?:";
    const char       *modificator = NULL;
    recobe_conf      *cf = &recobe_cf;
    recobe_chain     *chain0 = NULL;
    recobe_chain     *chain = NULL;
    const char       *kflag = NULL;
    const char       *cflag = NULL;
    const char       *tflag = NULL;
    const char       *Tflag = NULL;
    int               ex = EX_OK;
    int               code;
    int               fdou;
    babolo_opts      *bos;
    u_char           *z;
    const char       *o;
    int               c;

    bzero(cf, sizeof(recobe_conf));
    blin_ctl(BLIN_CTL_FSET, (cf->flags = BLIN_BIT0));
#   define blin_internal_flags (cf->flags)
    cf->bsubst = (u_char*)"?";
    if  (!(bos = babolo_openopts(Bpars_ESYM | Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (!!babolo_setopts(bos, Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    code = 0;
    for (tflag = 0;;) {
        c = babolo_getopts(bos);
        if  (!c) break;
        switch (c) {
        case 'b': cf->bsubst = (u_char*)babolo_getoptsarg(bos);
                ; break;
        case 'c': cflag = babolo_getoptsarg(bos);
                ; chain = recobe_openc(cf, cflag, 0, chain);
                ; if  (!chain0) chain0 = chain;
                ; break;
        case 'e': cf->esubst = (u_char*)babolo_getoptsarg(bos);
                ; break;
        case 'f': cf->flags |= RECOBE_FULLCHECK;
                ; break;
        case 'H': cf->flags |= RECOBE_ERHEXPRT;
                ; break;
        case 'k': kflag = babolo_getoptsarg(bos);
                ; chain = recobe_openk(cf, kflag, chain, modificator);
                ; if  (!chain0) chain0 = chain;
                ; modificator = NULL;
                ; cf->flags &= ~(RECOBE_USES | RECOBE_USE8 | RECOBE_USE9);
                ; break;
        case 'l': cf->flags |= RECOBE_PRELINERA;
                ; break;
        case 'm': modificator = babolo_getoptsarg(bos);
                ; break;
        case 'M': cf->flags |= (*babolo_getoptsarg(bos) & 0x00FF) << 8;
                ; break;
        case 'o': o = babolo_getoptsarg(bos);
                ; switch(o[0]) {
                  case 'b':
                  case 'l':
                  case 'B':
                  case 'L':
                      cf->flags |= o[0] & RECOBE_ORDER;
                      break;
                  default:
                      ifBLIN_QX0("Illegal argument %s of -o option", o);
                  }
                ; break;
        case 'p': cf->flags |= RECOBE_PRINTABLE;
                ; break;
        case 'P': cf->flags |= RECOBE_PRINTATLE;
                ; break;
        case 'q': cf->flags &= ~BLIN_MASK;
                ; break;
        case 's': code = babolo_getoptsnum(bos, 0);
                ; break;
        case 'S': cf->flags |= RECOBE_STRICT;
                ; break;
        case 't': tflag = babolo_getoptsarg(bos);
                ; break;
        case 'T': cf->flags |= RECOBE_TRACE;
                ; break;
        case 'v': BLIN_VERBOSE(cf->flags);
                ; break;
        case 'w': cf->flags |= RECOBE_USE8;
                ; break;
        case 'W': cf->flags |= RECOBE_USE9;
                ; break;
        case '=': blin_ctl(BLIN_CTL_FEQU, cf->flags);
                  switch((Tflag = babolo_getoptsarg(bos))[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);      exit(EX_OK);
                  case '1': babolo_optest(flar);          exit(EX_OK);
                  case '2': babolo_optext(bos);           exit(EX_OK);
                  case '3': babolo_dumpopts(bos);         exit(EX_OK);
                  }
                ; break;
        case '?': cflag = babolo_getoptsarg(bos);
                ; if  (!cflag) usage(EX_OK);
                ; switch (*cflag) {
                  case 'c': usagec(EX_OK);
                  case 's': usages(EX_OK);
                  case 'k': usagek(EX_OK);
                  default : usage(EX_USAGE);
                  }
        default : ifBLIN_QX0("Illegal flag -%c", bos->c);
                ; usage(EX_USAGE);
    }   }
    blin_ctl(BLIN_CTL_FEQU, cf->flags);
    if  (Tflag) {
        switch (Tflag[0]) {
        case '4': ex = recobe_TestOrRevert(++Tflag); break;
        case '5': ex = recobe_PrintAndMake(++Tflag); break;
        case 'S': ex = recobe_WriteAndMake(++Tflag); break;
        case '6': ex = recobe_PrintNoRecod(++Tflag); break;
        case '7': ex = recobe_G711DiffConv(++Tflag); break;
        case '8': ex = recobe_PrintFilling(++Tflag); break;
        case '9': ex = recobe_InterMediaTo(++Tflag); break;
        case 'E': ex = recobe_EtalonCompar(++Tflag); break;
        case 'R': ex = recobe_ReCalculator(++Tflag); break;
        case 'C': ex = recobe_ReCoordinate(++Tflag); break;
        case 'W': ex = recobe_PrintWideStr(++Tflag); break;
        }
    } else if (!!code) {
        ssize_t           hexlen = 0;
        if  (!(z = babolo_hextrin(babolo_getargs(bos), &hexlen))) {
            ifBLIN_QW0("babolo_hextrin");
            ex = EX_IOERR;
            goto out;
        }
        ifBLIN_QO1 blin_dumb(1, z, hexlen);
        if  (0 > code) hexlen = -hexlen;
        switch ((0 > code) ? -code : code) {
        case 1: printf("csum1 = %02X\n", recobe_csum1(z, hexlen));
                break;
        case 2: printf("crc24 = %06X\n", recobe_crc24(z, hexlen));
                break;
        case 3: printf("crc32 = %08X\n", recobe_crc32(z, hexlen));
                break;
        default:
                ifBLIN_QX0("Illegal code %d", code);
        }
    } else if (!!chain0) {
        chain0->next = recobe_openfile(cf, babolo_getargs(bos), RECOBE_PREFLEN);
        fdou = fileno(stdout);
        if  ((o = babolo_getargs(bos)) && !(fdou = open(o, O_WRONLY | O_TRUNC | O_CREAT, 0777))) {
            ifBLIN_QW0("fopen ou %s", o);
            ex = EX_IOERR;
            goto out;
        }
        recobe_chainout(chain, fdou); 
        mife_fini(chain0->next->converdata);
        close(fdou);
    } else {
        ifBLIN_QX0("No recoding");
    }
out:
    exit(ex);
#   undef blin_internal_flags
}
